//
//  MixMaster.cpp
//  SSS
//
//  Created by Jeffrey Harriman on 3/9/12.
//  Copyright (c) 2012 Stanford University. All rights reserved.
//

#include <iostream>
#include "MixMaster.h"

MixMaster::MixMaster(){
    // initialize mixing buffer array
    // requested frame size x2 should be enough
    m_mixBufArray = new SAMPLE * [NUM_TRACKS];
    for (int i = 0; i < NUM_TRACKS; i++){
        m_mixBufArray[i] = new SAMPLE[NUM_CHANNELS * FRAMESIZE * 2 ];  
    }
    // metronome
    m_metronome = new Metro();
    
    // init tracks
    m_track = new Track * [NUM_TRACKS];
    for (int i = 0; i < NUM_TRACKS; i++){
        m_track[i] = new Track();
    }

    m_track[AUDIO_TRACK] = new AudioRecord();
    m_track[AUDIO_TRACK_2] = new AudioRecord();
    m_track[LEAD_TRACK] = new LeadTrack();
    m_track[BASS_TRACK] = new BassTrack();
    m_track[CHORD_TRACK] = new ChordTrack();
    
    for (int i = 0; i < NUM_TRACKS; i++){
        m_track[i]->assignMetro(m_metronome); 
        m_track[i]->setLoopLength(m_metronome->getLoopLength());
    }

    
    // put all tracks in standby
    Stop();
}

void MixMaster::tick(Float32 *buffer, int numFrames ){
    // clear out the mix buffer 
    for (int i = 0; i < NUM_TRACKS; i++){
        // put the incoming buffer into the Audio Channel of the mixer
        for (int j = 0; j<numFrames; j++){
            if ((i == AUDIO_TRACK || i == AUDIO_TRACK_2) && !m_metronome->getCountIn()){
                m_mixBufArray[i][j*NUM_CHANNELS] = buffer[ NUM_CHANNELS * j ];
                
            }
            else
                m_mixBufArray[i][j*NUM_CHANNELS] = m_mixBufArray[i][j*NUM_CHANNELS + 1] = 0;
        }
    }
    
    // now clear the original buffer
    for (int i = 0; i < numFrames; i++){
        buffer[NUM_CHANNELS * i] = buffer[NUM_CHANNELS * i + 1] = 0.0;
    }
    
    if(!m_metronome->getCountIn()){
        // Audio Tracks
        if (m_track[AUDIO_TRACK]->m_active){
            ((AudioRecord *) m_track[AUDIO_TRACK])->tick( m_mixBufArray[AUDIO_TRACK], numFrames );
        }
        // Audio 2
        if (m_track[AUDIO_TRACK_2]->m_active){
            ((AudioRecord *) m_track[AUDIO_TRACK_2])->tick( m_mixBufArray[AUDIO_TRACK_2], numFrames );
        }
        
        // if on an audio track - fill the draw buffer
        if (*m_currentTrack == AUDIO_TRACK || *m_currentTrack == AUDIO_TRACK_2)
        {
            SAMPLE width = 1.5;
            SAMPLE x = 0;
            SAMPLE inc = width / numFrames;
            for( int i = 0; i < numFrames; i++ )
            {
                // x coordinate
                m_vertices[2*i] = x; x += inc;
                // y coordinate
                m_vertices[2*i+1] = 3 * m_mixBufArray[*m_currentTrack][i] ;
            
            }
        }
        
        for (int i = 0; i < numFrames; i++){
            // Lead Track
            if (m_track[LEAD_TRACK]->m_active){
                
                m_mixBufArray[LEAD_TRACK][2*i] = m_mixBufArray[LEAD_TRACK][2*i+1] = .7 * ((LeadTrack *) m_track[LEAD_TRACK])->tick(i);
            }
            
            // Bass Track
            if (m_track[BASS_TRACK]->m_active){
                m_mixBufArray[BASS_TRACK][2*i] = m_mixBufArray[BASS_TRACK][2*i+1] = .5 * ((BassTrack *) m_track[BASS_TRACK])->tick(i);
            }
        
            // Chord Track
            if (m_track[CHORD_TRACK]->m_active){
                m_mixBufArray[CHORD_TRACK][2*i] = m_mixBufArray[CHORD_TRACK][2*i+1] = .7 * ((ChordTrack *) m_track[CHORD_TRACK])->tick(i);
            }
        }
    }
    
    // Metronome Track (tick last)
    bool metro_active = false;
    for (int i = 0; i < NUM_TRACKS; i++){
        int track_mode = m_track[i]->getPlayMode();
        if (track_mode == RECORD || track_mode == PLAY || track_mode == RECORD_ARTICULATION){
            metro_active = true;
            break;
        }
    }
    
        
    if (metro_active){
        for (int i = 0; i < numFrames; i++){
            m_mixBufArray[METRO_TRACK][2*i] = m_mixBufArray[METRO_TRACK][2*i+1] = m_metronome->tick();
        }
    }
    
    
    // MIX!
    for (int i = 0; i < NUM_TRACKS; i++){
        for (int j = 0; j < numFrames; j++){
            // mono
            buffer[ j * NUM_CHANNELS] = buffer[ NUM_CHANNELS * j + 1 ] += 0.6 * m_mixBufArray[i][j*NUM_CHANNELS];
        }
    }
    
    
}

// send keyOn messages based on current active control track
void MixMaster::keyOn(){
    m_track[*m_currentTrack]->keyOn();
}

// send keyOff messages based on current active control track
void MixMaster::keyOff(){
    m_track[*m_currentTrack]->keyOff();
}

// send note to appropriate instrument
void MixMaster::setNote(int note){
    if (*m_currentTrack == LEAD_TRACK || *m_currentTrack == BASS_TRACK)
        ((MonophonicTrack *) m_track[*m_currentTrack])->setNote(note);
    else if (*m_currentTrack == CHORD_TRACK)
        ((ChordTrack *) m_track[CHORD_TRACK])->setScaleDegree(note);
}

// //////////////////////////////////
// GUI functions
//////////////////////////////////
// Butons
void MixMaster::Record(){
    m_metronome->setActive(true);
    m_metronome->startCountIn();
    //  set all tracks to play
    for (int i = 0; i < NUM_TRACKS; i++){
        if (m_track[i]->m_active)
            m_track[i]->setPlayMode(PLAY);
    }
    // set current track to record
    m_track[*m_currentTrack]->setPlayMode(RECORD);

}
void MixMaster::Play(){
    m_metronome->setActive(false);
    // all play
    for (int i = 0; i < NUM_TRACKS; i++){
        if (m_track[i]->m_active)
            m_track[i]->setPlayMode(PLAY);
    }
}
void MixMaster::Stop(){
    m_metronome->stop();
    for (int i = 0; i < NUM_TRACKS; i++){
        m_track[i]->setPlayMode(STANDBY);
        m_track[i]->keyOff();
    }
}
void MixMaster::Dub(){
    m_metronome->setActive(true);
    m_metronome->startCountIn();
    //  set all tracks to play
    for (int i = 0; i < NUM_TRACKS; i++){
        if (m_track[i]->m_active)
            m_track[i]->setPlayMode(PLAY);
    }
    // set current track to record
    m_track[*m_currentTrack]->setPlayMode(RECORD_ARTICULATION);

}

void MixMaster::Erase(){

}

// Switches
void MixMaster::setPlayArticulation(bool value){
    
}

void MixMaster::setActive(bool value){
     
}

//////////////
//  fetching, loading track data
/////////////
audioRecording MixMaster::getAudioRecording(int track){
    audioRecording theRecording;
    theRecording.length = 0;
    
    if (track == AUDIO_TRACK || track == AUDIO_TRACK_2)
        theRecording = ((AudioRecord *) m_track[track])->getAudioRecording();
    
    return theRecording;
};

int MixMaster::loadAudioRecording(audioRecording recording){
    if (*m_currentTrack == AUDIO_TRACK || *m_currentTrack == AUDIO_TRACK_2){
        ((AudioRecord*) m_track[*m_currentTrack])->loadAudioRecording(recording);
        return 1;
    }
    else
        return -1;
}

int MixMaster::loadAudioRecording(int track, audioRecording recording){
    if (track == AUDIO_TRACK || track == AUDIO_TRACK_2){
        ((AudioRecord*) m_track[track])->loadAudioRecording(recording);
        return 1;
    }
    else
        return -1;
}

synthRecording MixMaster::getNoteRecording(int track){
    synthRecording theRecording;
    theRecording.length = 0;

    
    if (track == BASS_TRACK || track == LEAD_TRACK || track == CHORD_TRACK)
        theRecording = (m_track[track])->getNoteRecording();
    
    return theRecording;
};

artRecording MixMaster::getArtRecording(int track){
    artRecording recording;
    recording.length = 0;
    
    recording = m_track[track]->getArtRecording();
    return recording;
}

int MixMaster::loadNoteRecording(synthRecording recording){
    if (*m_currentTrack == CHORD_TRACK || *m_currentTrack == BASS_TRACK || *m_currentTrack == LEAD_TRACK){
        (m_track[*m_currentTrack])->loadNoteRecording(recording);
        return 1;
    }
    else
        return -1;
}

int MixMaster::loadNoteRecording(int track, synthRecording recording){
    if (track == CHORD_TRACK || track == BASS_TRACK || track == LEAD_TRACK){
        (m_track[track])->loadNoteRecording(recording);
        return 1;
    }
    else
        return -1;
}

int MixMaster::loadArtRecording(artRecording recording){
    
    (m_track[*m_currentTrack])->loadArtRecording(recording);
    return 1;
   
}

int MixMaster::loadArtRecording(int track, artRecording recording){
    
    (m_track[track])->loadArtRecording(recording);
    return 1;
    
}