//
//  Track.cpp
//  SSS
//
//  Created by Jeffrey Harriman on 3/8/12.
//  Copyright (c) 2012 Stanford University. All rights reserved.
//

#include <iostream>
#include "Track.h"

Track::Track(){
    m_active = true;
    setPlayMode(STANDBY);
    m_maxLength = 300 * JIF_SRATE; // 5 minutes!
    m_gain = .5;
    // envelope
    m_adsr = new stk::ADSR();
    m_adsr->keyOff();

    m_artRec = new recordData[1];
    m_artRec[0].type = 0;
    m_artRec[0].value = 0;
    m_artRec[0].when.measure = 0;
    m_artRec[0].when.beat = 0;
    m_artRec[0].when.subBeat = 0;
    
    m_noteRec = new recordData[1];
    m_noteRec[0].type = 0;
    m_noteRec[0].value = 0;
    m_noteRec[0].when.measure = 0;
    m_noteRec[0].when.beat = 0;
    m_noteRec[0].when.subBeat = 0;
    
    m_artPtr = m_numArts = 0;
    m_notePtr = m_numNotes = 0;
    
    //
    m_key = 0;
    m_keyName = -1;
}

Track::~Track(){
    delete m_adsr;
}

void Track::setLoopLength(int measures, int beatsPer, float tempo){
    int numFrames = JIF_SRATE * measures * beatsPer * tempo/60;
    setLoopLength(numFrames);
}

void Track::setLoopLength(int numFrames){
    if (numFrames <= m_maxLength)
        m_length = numFrames;
    else
        m_length = m_maxLength;
}

//-------------------------
// setPlayMode()
// update mode for this track, if record reset the appropriate num wrappers
//-------------------------
void Track::setPlayMode(int mode){
    m_playMode = mode;
    setTimeZero();
    
    if (m_playMode == RECORD){
        m_numNotes = 0;
        m_numArts = 0;
        // set to -1 so first note change will be recorded even if its the same as current note...
        m_keyName = -1;
    }
    
    if (m_playMode == RECORD_ARTICULATION)
        m_numArts = 0;
}

void Track::setTimeZero(){
    m_ptr = 0;
    m_artPtr = 0;
    m_notePtr = 0;
}

void Track::keyOn(){
    if (m_active){
        if (m_playMode == STANDBY || m_playMode == PLAY || m_playMode == LOOP || m_metronome->m_countIn)
            m_adsr->keyOn();
        else if ((m_playMode == RECORD || m_playMode == RECORD_ARTICULATION) && !m_metronome->m_countIn){
            m_adsr->keyOn();
            BEAT_COUNT timeStamp = m_metronome->getQuantizedBeat();
            recordData new_record;
            new_record.when = timeStamp;
            new_record.type = KEY_ON;
            recordArticulation(new_record);
        }
    }
}

void Track::keyOff(){
    if (m_active){
        if (m_playMode == STANDBY || m_playMode == PLAY || m_playMode == LOOP || m_metronome->m_countIn)
            m_adsr->keyOff();
        else if ((m_playMode == RECORD || m_playMode == RECORD_ARTICULATION) && !m_metronome->m_countIn){
            m_adsr->keyOff();
            BEAT_COUNT timeStamp = m_metronome->getQuantizedBeat();
            recordData new_record;
            new_record.when = timeStamp;
            new_record.type = KEY_OFF;
            recordArticulation(new_record);
        }
    }
}

void Track::recordNote(recordData dataIn){
    if (!m_metronome->m_countIn){
        m_numNotes++;
        recordData * temp;
        temp = new recordData[m_numNotes];
        
        // copy old data
        for (int i = 0; i < m_numNotes-1; i++){
            temp[i] = m_noteRec[i];
        }

        temp[m_notePtr] = dataIn;
        //NSLog(@"recorded note: %i, at: m %i, b %i, s: %i", (int)m_noteRec[m_notePtr].value, m_noteRec[m_notePtr].when.measure, m_noteRec[m_notePtr].when.beat, m_noteRec[m_notePtr].when.subBeat);
        // point at the new data
        m_noteRec = temp;
        m_notePtr++;
    }
}


void Track::recordArticulation(recordData dataIn){
    m_numArts++;
    recordData * temp;
    temp = new recordData[m_numArts];

    // copy old data
    for (int i = 0; i < m_numArts-1; i++){
        temp[i] = m_artRec[i];
    }
    
    
    temp[m_artPtr] = dataIn;
    // point at the new data
    m_artRec = temp;
    m_artPtr++;
}

void Track::playArticulation(){
    int initArtPtr = m_artPtr;
    // while we have new articulation data
    BEAT_COUNT currentBeat =m_metronome->getQuantizedBeat();
    BEAT_COUNT nextArt = m_artRec[m_artPtr].when;
//    NSLog(@"mtro: m: %i,b: %i,s: %i", currentBeat.measure, currentBeat.beat, currentBeat.subBeat);
//    NSLog(@"next: m: %i,b: %i,s: %i", nextArt.measure, nextArt.beat, nextArt.subBeat);
    while ((nextArt.measure == currentBeat.measure) && ( nextArt.beat == currentBeat.beat ) && ( nextArt.subBeat == currentBeat.subBeat ) )
    {
        if(m_artRec[m_artPtr].type == KEY_ON)
            keyOn();
               
        if(m_artRec[m_artPtr].type == KEY_OFF)
            keyOff();
               
        m_artPtr++;
        m_artPtr %= m_numArts;
        nextArt = m_artRec[m_artPtr].when;
        // if there's only one articulation pointer don't keep trying to read it
        if (initArtPtr == m_artPtr)
            break;
    }
}

synthRecording Track::getNoteRecording(){
    synthRecording recording;
    recording.length = m_numNotes;
    recording.note_data = m_noteRec;
    
    return recording;
}

artRecording Track::getArtRecording(){
    artRecording recording;
    recording.length = m_numArts;
    recording.art_data = m_artRec;
    
    return recording;
}

void Track::loadNoteRecording(synthRecording recording){
    m_numNotes = recording.length;
    
    // clear the old, make way for the new
    delete m_noteRec;
    m_noteRec = new recordData[m_numNotes];
    // copy the buffer
    for (int i = 0; i < m_numNotes; i++){
        m_noteRec[i] = recording.note_data[i];
    }
    
    setPlayMode(STANDBY);
}

void Track::loadArtRecording(artRecording recording){
    m_numArts = recording.length;
    
    // clear the old, make way for the new
    delete m_artRec;
    m_artRec = new recordData[m_numArts];
    
    // copy
    for (int i = 0; i < m_numArts; i++){
        m_artRec[i] = recording.art_data[i];
    }
    
    setPlayMode(STANDBY);
    
}

